#!/usr/bin/python

#=============================================================================================
# A module implementing multistate Bennett acceptance ratio (MBAR) for the analysis
# of multiple simulations conducted under arbitrary conditions.
#
# This module provides a class for computing free energy differences and expectations of
# observables for arbitrary thermodynamic states given equilibrium data collected from multiple states.
#
# Reference:
#
# Shirts MR and Chodera JD. Statistically optimal analysis of multiple equilibrium simulations.
# Submitted, 2007.
#=============================================================================================

#=============================================================================================
# COPYRIGHT NOTICE
#
# Written by John D. Chodera <jchodera@gmail.com>, Dill lab, UCSF, 2007.
#
# Copyright (c) 2007 The Regents of the University of California.  All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify it under the terms of
# the GNU General Public License as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with this program;
# if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
# Boston, MA  02110-1301, USA.
#=============================================================================================

#=============================================================================================
# TODO
# * L-BFGS
# * Double-check correspondence of comments to equation numbers once manuscript has been finalized.
# * Speed things up with compiled code (using f2py and SciPy?).
# The only thing that requires acceleration is the computation of the log weights.
#=============================================================================================

#=============================================================================================
# CHAGELOG
# 2007-07-29 JDC
# * Added changelog.
# * Added preliminary Newton-Raphson implementation.  Needs tuning and more documentation.
# Currently, the first step is scaled by gamma = 0.1, and subsequent steps are unscaled, but
# quadratically-convergent behavior is only seen after a few steps.  L-BFGS should help immensely.
# * Split out self-consistent iteration code into separate function
# * Added new uncertainty method 'svd-ew' where SVD is avoided by used of eigenvalue decomposition of W'W.
# This saves a great deal of memory and time for cases when N is large.  Tested by comparing Theta against 'svd' method directly.
# This method is now the default.
# * Changed the default options for 'uncertainty_method' for several methods to be None, which now defaults to 'svd-ew' when covariance is computed.
# 2007-07-31 JDC
# * Added option to MLWHAM constructor to allow choice of methods between self-consistent iteration and Newton-Raphson.
# * Added more documentation on Newton-Raphson method.
# 2007-07-31 JDC
# * Enhanced stability of 'svd-ew' method for computing covariance matrix by setting small negative eigenvalues S2 to zero.
# * Solved problem with Newton-Raphson and states with no samples by substituting pseudoinverse for inverse of Hessian.
# * Sometimes, Newton-Raphson would produce small differences in free energies of states that were the same, so added final update of free energies.
# 2007-08-16 JDC
# * Minor changes to address a bug Guha observed with wacky data:
# * Changed energy storage to float64 (from float32)
# * Added an iteration of self-consistent-iteration in front of Newton-Raphson to reduce convergence problems with wacky data.
# 2007-08-17 JDC
# * Changed the default behavior of subsampleCorrelatedData to ensure we end up with closer to ~ N/g samples. Old behavior is still available with optional flag.
# * Fixed example code in class comments.
# 2007-10-28 MRS
# * Changed from MLWHAM name to MBAR
# * Made some of the defaults consistent.
#=============================================================================================

#=============================================================================================
# Version CONTROL INFORMATION
__version__ = "$Revision: 1.1 $"
# $Source: /share_kubo/Storage/cvsroot/MLWHAM/pyMBAR/MBAR.py,v $
#=============================================================================================

#=============================================================================================
# IMPORTS
from numpy import * # matrix methods
from numpy.linalg import * # linear algebra methods
#=============================================================================================

#=============================================================================================
# Exception class definition
#=============================================================================================
class ParameterError(Exception):
  """An error in the input parameters has been detected.

  """

#=============================================================================================
# Methods used by the MBAR class that can be of general use.
#=============================================================================================
def statisticalInefficiency(A_n, B_n, fast = False):
  """Compute the (cross) statistical inefficiency of (two) timeseries.

  REQUIRED ARGUMENTS  
    A_n[n] is nth value of timeseries A.  Length is deduced from vector.
    B_n[n] is nth value of timeseries B.  Length is deduced from vector.

  OPTIONAL ARGUMENTS  
    fast can be set to True to give a less accurate but very quick estimate (default False)

  RETURNS
    g is the statistical inefficiency (equal to 1 + 2 tau, where tau is the correlation time).

  NOTES 
    The same timeseries can be used for both A_n and B_n to get the autocorrelation statistical inefficiency.
    The fast method described in Ref [1] is used to compute g.

  REFERENCES  
    [1] J. D. Chodera, W. C. Swope, J. W. Pitera, C. Seok, and K. A. Dill. Use of the weighted
    histogram analysis method for the analysis of simulated and parallel tempering simulations.
    JCTC 3(1):26-41, 2007.

  """

  # Create numpy copies of input arguments.
  A_n = array(A_n)
  B_n = array(B_n)

  # Get the length of the timeseries.
  N = A_n.size

  # Be sure A_n and B_n have the same dimensions.
  if(A_n.shape != B_n.shape):
    raise ParameterError('A_n and B_n must have same dimensions.')

  # Initialize statistical inefficiency estimate with uncorrelated value.
  g = 1.0
    
  # Compute means and variance.
  mu_A = A_n.mean()
  mu_B = B_n.mean()

  # Make temporary copies at high precision with means subtracted off.
  dA_n = A_n.astype(float64) - mu_A
  dB_n = B_n.astype(float64) - mu_B

  # sigma2_AB = sum((A_n-mu_A) * (B_n-mu_B)) / (float(N)-1.0) # unbiased estimator
  sigma2_AB = (dA_n * dB_n).mean() # standard estimator to ensure C(0) = 1
  if(sigma2_AB == 0):
    raise ParameterException('Sample covariance sigma_AB^2 = 0 -- cannot compute statistical inefficiency')

  # Accumulate the integrated correlation time by computing the normalized correlation time at
  # increasing values of t.  Stop accumulating if the correlation function goes negative, since
  # this is unlikely to occur unless the correlation function has decayed to the point where it
  # is dominated by noise and indistinguishable from zero.
  t = 1
  increment = 1
  while (t < N-1):

    # compute normalized fluctuation correlation function at time t
    C = sum( dA_n[0:(N-t)]*dB_n[t:N] + dB_n[0:(N-t)]*dA_n[t:N] ) / (2.0 * float(N-t) * sigma2_AB)
    
    # Terminate if the correlation function has crossed zero.  
    if (C <= 0.0):
      break

    # Accumulate contribution to the statistical inefficiency.
    g += 2.0 * C * (1.0 - float(t)/float(N)) * float(increment)

    # Increment t and the amount by which we increment t.
    t += increment

    # Increase the interval if "fast mode" is on.
    if fast: increment += 1

  # Return the computed statistical inefficiency.
  return g

def subsampleCorrelatedData(A_t, fast = False, conservative = False):
  """Determine the indices of an uncorrelated subsample of the data.

  REQUIRED ARGUMENTS  
    A_t (T array) - A_t[t] is the t-th value of timeseries A(t).  Length is deduced from vector.

  OPTIONAL ARGUMENTS  
    fast (logical) - fast can be set to True to give a less accurate but very quick estimate (default False)
    conservative (logical) - if set to True, uniformly-spaced indices are chosen with interval ceil(g), where
      g is the statistical inefficiency.  Otherwise, indices are chosen non-uniformly with interval of
      approximately g in order to end up with approximately T/g total indices

  RETURNS  
    indices (list of int) - the indices of an uncorrelated subsample of the data

  NOTES
    The statistical inefficiency is computed with the function computeStatisticalInefficiency().

  TODO
    Instead of using regular stride, use irregular stride so more data can be fit in when g is non-integral.

  EXAMPLE

  # Suppose our timeseries is A_t[t], t = 0..T
  indices = subsampleCorrelatedData(A_t) # determine indices of uncorrelated samples
  A_n = A_t[indices] # construct a set of uncorrelated samples

  """

  # Create numpy copy of arrays.
  A_t = array(A_t)

  # Get the length of the timeseries.
  T = A_t.size

  # Compute the statistical inefficiency for the timeseries.
  g = statisticalInefficiency(A_t, A_t, fast = fast)

  if conservative:
    # Round g up to determine the stride we can use to pick out regularly-spaced uncorrelated samples.
    import math
    stride = int(math.ceil(g))
    print "g = %f, using stride of %d" % (g, stride)
    
    # Assemble list of indices of uncorrelated snapshots.
    indices = range(0, T, stride)
  else:
    # Choose indices as floor(n*g), with n = 0,1,2,..., until we run out of data.
    import math
    indices = []
    n = 0
    while int(math.floor(n*g)) < T:
      indices.append(int(math.floor(n*g)))
      n += 1

  # Return the list of indices of uncorrelated snapshots.
  return indices

#=============================================================================================
# MBAR class definition
#=============================================================================================
class MBAR:
  """Maximum-likelihood weighted histogram analysis method without histograms (MBAR) for the analysis of multiple equilibrium simulations.

  NOTES
    Note that this method assumes the data are uncorrelated.  The data must be subsampled to produce independent samples if the data is correlated (see example below).

  REFERENCE
    [1] Shirts MR and Chodera JD. Statistically optimal analysis of multiple equilibrium simulations. Submitted, 2007.

  EXAMPLES

  * Example 1. Computation of relative free energies from an alchemical simulation.

  # Import the MBAR analysis code.
  import MBAR

  # Suppose the energies sampled from each simulation are u_klt, where u_klt[k,l,t] is the reduced potential energy
  # of snapshot t \in 1,...,T_k of simulation k \in 1,...,K evaluated at reduced potential for state l.

  # First, we subsample the data to obtain statistically uncorrelated samples.
  N_k = zeros([K], int32) # N_k[k] will denote the number of correlated snapshots from state k
  u_kln = zeros([K, K, T_k.max()], float64) # u_kln[k,l,n] will store the reduced potential energy at state l of uncorrelated snapshot n \in 1..N_k[k] from state k.
  for k in range(0,K):
      # Determined indices of statistically independent configurations by analyzing the correlation structure of the timeseries data.
      indices = MBAR.subsampleCorrelatedData(u_klt[k,k,0:T_k[k]])

      # Subsample data.
      N_k[k] = len(indices)
      for l in range(0,K):
          u_kln[k,l,0:N_k[k]] = u_klt[k,l,indices]

  # Initialize MBAR with reduced energies u_kln and number of uncorrelated configurations from each state N_k.
  # 
  # u_kln[k,l,n] is the reduced potential energy beta*U_l(x_kn), where U_l(x) is the potential energy function for state l,
  # beta is the inverse temperature, and and x_kn denotes uncorrelated configuration n from state k.
  #
  # N_k[k] is the number of configurations from state k stored in u_knm
  # 
  # Note that this step may take some time, as the relative dimensionless free energies f_k are determined at this point.
  mbar = MBAR.mbar(u_kln, N_k)

  # Extract dimensionless free energy differences and their statistical uncertainties.
  (Deltaf_ij, dDeltaf_ij) = mbar.getFreeEnergyDifferences()
  print 'Unit-bearing free energy difference between states 1 and K: %f +- %f' % ( (1./beta) * Deltaf_ij[0,K-1], (1./beta) * dDeltaf_ij[0,K-1])

  # Compute the expectation of some observable A(x) at each state i, and associated uncertainty matrix.
  # Here, A_kn[k,n] = A(x_{kn})
  (A_k, dA_k) = mbar.computeExpectation(A_kn)
    
  """

  #=============================================================================================
  # Constructor.
  #=============================================================================================
  def __init__(self, u_kln, N_k, maximumIterations = 1000, relative_tolerance = 1.0e-6, verbose = False, initial_f_k = None, method = 'Newton-Raphson'):
    """Initialize multistate Bennett acceptance ratio (MBAR) on a set of simulation data.

    Upon initialization, the dimensionless free energies for all states are computed.
    This may take anywhere from seconds to minutes, depending upon the quantity of data.

    After initialization, the computed free energies may be obtained by a call to 'getFreeEnergies()', or
    free energies or expectation at any state of interest can be computed by calls to 'computeFreeEnergy()' or
    'computeExpectation()'.
    
    REQUIRED ARGUMENTS
      u_kln (KxKxNmax float array) - u_kln[k,l,n] is the reduced potential energy of uncorrelated configuration n sampled from state k, evaluated at state l
      N_k (K int array) - N_k[k] is the number of uncorrelated snapshots sampled from state k -- this can be zero if the expectation or free energy
          of this state is desired but no samples were drawn from this state

    NOTES
      The reduced potential energy u_kln[k,l,n] = u_l(x_{kn}), where the reduced potential energy u_l(x) is defined (as in the text) by:
        u_l(x) = beta_l [ U_l(x) + p_l V(x) + mu_l' n(x) ]
      where
        beta_l = 1/(kB T_l) is the inverse temperature of condition l, where kB is Boltzmann's constant
        U_l(x) is the potential energy function for state l
        p_l is the pressure at state l (if an isobaric ensemble is specified)
        V(x) is the volume of configuration x
        mu_l is the M-vector of chemical potentials for the various species, if a (semi)grand ensemble is specified, and ' denotes transpose
        n(x) is the M-vector of numbers of the various molecular species for configuration x, corresponding to the chemical potential components of mu_m.

      The configurations x_kn must be uncorrelated.  This can be ensured by subsampling a correlated timeseries with a period larger than the statistical inefficiency,
      which can be estimated from the potential energy timeseries {u_k(x_kn)}_{n=1}^{N_k} using the provided utility function 'statisticalInefficiency()'.
      See the help for this function for more information.

    OPTIONAL ARGUMENTS
      maximumIterations (int) - can be set to limit the maximum number of iterations performed (default 1000)
      relative_tolerance (int) - can be set to determine the relative tolerance convergence criteria (defailt 1.0e-6)
      verbosity (logical) - should be set to True if verbose debug output is desired (default False)
      initial_f_k (numpy K float64 array) - should be set to a numarray K-array with initial dimensionless free energies to use as a guess (default None, which sets all f_k = 0)
      method (string) - choose method for determination of dimensionless free energies: 'self-consistent-iteration' or 'Newton-Raphson' (default: 'Newton-Raphson')
        Newton-Raphson starts with one iteration of self-consistent-iteration as a safeguard in case initial guess at f_k is in a highly non-quadratic region        

    """

    # Store local copies of necessary data.
    self.N_k = array(N_k, dtype = int32) # N_k[k] is the number of configurations generated from state k
    self.u_kln = array(u_kln, dtype = float64) # u_kln[k,l,n] is the reduced potential energy of configuration n from state k evaluated at u_l(x) 

    # Gets dimensions of reduced energy matrix.
    [K, L, N_max] = self.u_kln.shape    

    # Perform consistency checks on dimensions.
    if K != L:
      raise ParameterError('u_kln[0:K, 0:L, 0:N_max] must have dimensions K == L.')
    if any(N_k > N_max):
      raise ParameterError('All N_k must be <= N_max, the third dimension of u_kln[0:K, 0:L, 0:N_max].')

    # Store local copies of other data
    self.K = K # number of thermodynamic states
    self.N_max = N_max # maximum number of configurations per state
    self.N = sum(self.N_k) # N = \sum_{k=1}^K N_k is the total number of uncorrelated configurations pooled across all states
    self.verbose = verbose # verbosity level -- if True, will print extra debug information
    
    # Initialize estimate of relative dimensionless free energy of each state to zero.
    # Note that f_k[0] will be constrained to be zero throughout.
    self.f_k = zeros([self.K], dtype = float64)

    # If an initial guess of the relative dimensionless free energies is specified, start with that.
    if initial_f_k != None:
      if self.verbose: print "WHAM: Initializing f_k with provided initial guess."
      # Cast to numpy array.
      initial_f_k = array(initial_f_k, dtype = float64)
      # Check shape
      if initial_f_k.shape != self.f_k.shape:
        raise ParameterError("initial_f_k must be a %d-dimensional numpy array." % self.K)
      # Initialize f_k with provided guess.
      self.f_k = initial_f_k
      if self.verbose: print self.f_k
      # Shift all free energies such that f_0 = 0.
      self.f_k[:] = self.f_k[:] - self.f_k[0]
      
    # Create a list of indices of all configurations in kn-indexing.
    mask_kn = zeros([self.K,self.N_max], dtype=bool_)
    for k in range(0,self.K):
      mask_kn[k,0:N_k[k]] = True
    # Create a list from this mask.
    self.indices = where(mask_kn)

    # Determine list of k indices for which N_k != 0
    self.nonzero_N_k_indices = where(self.N_k != 0)[0]

    if self.verbose:
      print "N_k = "
      print N_k

    # Determine dimensionles free energies.
    if method == 'self-consistent-iteration':
      # Use self-consistent iteration of MBAR equations.
      self.selfConsistentIteration(maximumIterations = maximumIterations, relative_tolerance = relative_tolerance, verbose = verbose)      
    elif method == 'Newton-Raphson':         
      # Use Newton-Raphson, starting with one self-consistent iteration.
      self.selfConsistentIteration(maximumIterations = 1000, relative_tolerance = relative_tolerance, verbose = verbose)            
      self.NewtonRaphson(first_gamma = 0.1, maximumIterations = maximumIterations, relative_tolerance = relative_tolerance, verbose = verbose)
    else:
      raise ParameterError("Specified method = '%s' is not a valid method. Specify either 'self-consistent-iteration' or 'Newton-Raphson'.")    

    # Print final dimensionless free energies.
    if self.verbose:
      print "Final dimensionless free energies"
      print "f_k = "
      print self.f_k      
        
    # Compute normalized weights.
    self.W_nk = zeros([self.N, self.K], dtype = float64)
    for l in range(0,K):
      # Compute unnormalized log weights.
      log_w_kn = self._computeUnnormalizedLogWeights(self.u_kln[:,l,:])      
      # Store normalized weights.
      self.W_nk[:,l] = exp(log_w_kn[self.indices] + self.f_k[l])

    # Compute the normalization constants c_k for the states.
    # Note that these are normally not used due to potential under/overflow errors.
    self.c_k = exp(-self.f_k)    

    return

  def getFreeEnergyDifferences(self, uncertainty_method = None, warning_cutoff = 1.0e-10):
    """Retrieve the dimensionless free energy differences and associated uncertainties among all thermodynamic states.

    REQUIRED ARGUMENTS

      none

    RETURN VALUES
    
      Deltaf_ij (numarray KxK Float matrix) - Deltaf_ij[i,j] = f_j - f_i, the dimensionless free energy difference between states i and j
      dDeltaf_ij (numarray KxK Float matrix) - dDeltaf_ij[i,j] is the estimated statistical uncertainty in Deltaf_ij[i,j]

    OPTIONAL ARGUMENTS
      uncertainty_method (string) - choice of method used to compute asymptotic covariance method, or None to use default
                        See help for computeAsymptoticCovarianceMatrix() for more information on various methods. (default: svd)
      warning_cutoff (float) - warn if squared-uncertainty is negative and larger in magnitude than this number (default: 1.0e-10)

    NOTES
      Computation of the covariance matrix may take some time.      
      
    """

    # Compute asymptotic covariance matrix using specified method.
    Theta_ij = self._computeAsymptoticCovarianceMatrix(self.W_nk, self.N_k, uncertainty_method)
        
    # Compute matrix of free energy differences between states and associated uncertainties.
    Deltaf_ij = zeros([self.K,self.K], dtype = float64)
    dDeltaf_ij = zeros([self.K,self.K], dtype = float64)
    for i in range(0, self.K):
      for j in range(0, self.K):
        if i != j:
          # Compute dimensionless free energy difference and associated uncertainty (Eq. 14 of [1]).
          Deltaf_ij[i,j] = (self.f_k[j] - self.f_k[i])

          # Compute associated squared-uncertainty (estimated variance of estimate of expectation).
          d2DeltaF = Theta_ij[i,i] + Theta_ij[j,j] - 2.0 * Theta_ij[i,j]           

          # Throw an error if squared uncertainty is large and negative -- otherwise, correct to zero.
          if (d2DeltaF < 0.0):
            if(-d2DeltaF > warning_cutoff):
              print "Squared uncertainty is negative.  d2DeltaF = %e" % d2DeltaF
            else:
              d2DeltaF = 0.0
          # Compute uncertainty from squared uncertainty.
          dDeltaf_ij[i,j] = math.sqrt( d2DeltaF )

    # Return matrix of free energy differences and uncertainties.
    return (Deltaf_ij, dDeltaf_ij)

  def computeExpectations(self, A_kn, uncertainty_method = None):
    """Compute the expectation of an observable of phase space function A(x) at all K states.

    REQUIRED ARGUMENTS
      A_kn (numarray KxN_max Float matrix) - A_kn[k,n] = A(x_kn)

    OPTINAL ARUMENTS
      uncertainty_method (string) - choice of method used to compute asymptotic covariance method, or None to use default
                        See help for computeAsymptoticCovarianceMatrix() for more information on various methods. (default: None)

    RETURN VALUES
      A_k (numarray K Float matrix) - A_k[k] is the estimate for the expectation of A(x) for state k.
      dA_k (numarray K float matrix) - dA_k[k] is uncertainty estimate for A_k[k]

    """

    # Convert to numpy matrix.
    A_kn = array(A_kn, dtype = float64)

    # Retrieve N and K for convenience.
    N = self.N
    K = self.K

    # Convert A_kn to n = 1..N indexing.
    A_n = A_kn[self.indices]    

    # Augment W_nk, N_k, and c_k for q_A(x) for the observable, with one extra row/column for each state.
    W_nk = zeros([N, K*2], dtype = float64) # weight matrix
    N_k = zeros([K*2], dtype = int32) # counts
    c_k = zeros([K*2], dtype = float64) # normalization constants    

    # Fill in first half of matrix with existing q_k(x) from states.
    W_nk[:,0:K] = self.W_nk
    N_k[0:K] = self.N_k
    c_k[0:K] = self.c_k

    # Compute the remaining rows/columns of W_nk and c_k for the observabls.
    # TODO: Make this more stable to under/overflow.
    for l in range(0,K):
      # Compute unnormalized weights.
      # A(x_n) q_k(x_n) / \sum_{k'=1}^K N_{k'} exp[f_{k'} - q_{k'}(x_n)]      
      W_nk[:,K+l] = A_n[:] * W_nk[:,l] * c_k[l]
      # Compute normalization constant from unnormalized weights
      c_k[K+l] = sum(W_nk[:,K+l])
      # Normalize weights
      W_nk[:,K+l] /= c_k[K+l]
           
    # Compute augmented asymptotic covariance matrix.
    Theta_ij = self._computeAsymptoticCovarianceMatrix(W_nk, N_k, method = uncertainty_method)

    # Compute estimators and uncertainties.
    A_k = zeros([K], dtype = float64)
    dA_k = zeros([K], dtype = float64)    
    for k in range(0,K):
      A_k[k] = sum(W_nk[:,k] * A_n[:])
      dA_k[k] = abs(A_k[k]) * sqrt(Theta_ij[K+k,K+k] + Theta_ij[k,k] - 2.0 * Theta_ij[k,K+k])
        
    # Return expectations and uncertainties.
    return (A_k, dA_k)

  #=============================================================================================
  # EXPERIMENTAL METHODS - USE AT YOUR OWN RISK!
  #=============================================================================================
  def computeEntropyAndEnthalpy(self, uncertainty_method = None):
    """Compute the decomposition of the free energy difference between states 1 and N into reduced free energy differences, reduced potential (enthalpy) differences, and reduced entropy (S/k) differences.

    OPTINAL ARUMENTS
      uncertainty_method (string) - choice of method used to compute asymptotic covariance method, or None to use default
                        See help for computeAsymptoticCovarianceMatrix() for more information on various methods. (default: None)

    RETURN VALUES
      Delta_f_ij (KxK numpy float matrix) - Delta_f_ij[i,j] is the dimensionless free energy difference f_j - f_i
      dDelta_f_ij (KxK numpy float matrix) - uncertainty in Delta_f_ij
      Delta_u_ij (KxK numpy float matrix) - Delta_u_ij[i,j] is the reduced potential energy difference u_j - u_i
      dDelta_u_ij (KxK numpy float matrix) - uncertainty in Delta_f_ij
      Delta_s_ij (KxK numpy float matrix) - Delta_s_ij[i,j] is the reduced entropy difference S/k between states i and j (s_j - s_i)
      dDelta_s_ij (KxK numpy float matrix) - uncertainty in Delta_s_ij

    """

    # Retrieve N and K for convenience.
    N = self.N
    K = self.K

    # Augment W_nk, N_k, and c_k for q_A(x) for the potential energies, with one extra row/column for each state.
    W_nk = zeros([N, K*2], dtype = float64) # weight matrix
    N_k = zeros([K*2], dtype = int32) # counts
    c_k = zeros([K*2], dtype = float64) # normalization constants    

    # Fill in first half of matrix with existing q_k(x) from states.
    W_nk[:,0:K] = self.W_nk
    N_k[0:K] = self.N_k
    c_k[0:K] = self.c_k

    # Compute the remaining rows/columns of W_nk and c_k for the potential energy observable.
    # TODO: Make this more stable to under/overflow.
    u_min = self.u_kln.min()
    print "u_min = "
    print u_min
    u_i = zeros([K], dtype = float64)    
    for l in range(0,K):
      # Convert potential energies to n = 1..N indexing.
      u_kn = self.u_kln[:,l,:] - u_min
      A_n = u_kn[self.indices]      
      # Compute unnormalized weights.
      # A(x_n) q_k(x_n) / \sum_{k'=1}^K N_{k'} exp[f_{k'} - q_{k'}(x_n)]      
      W_nk[:,K+l] = A_n[:] * W_nk[:,l] * c_k[l]
      # Compute normalization constant from unnormalized weights
      c_k[K+l] = sum(W_nk[:,K+l])
      # Normalize weights
      W_nk[:,K+l] /= c_k[K+l]
      # Compute expectations of energies
      u_i[l] = c_k[K+l] / c_k[l]

    print "W_nk ="
    print W_nk

    print "u_i = "
    print u_i
           
    # Compute augmented asymptotic covariance matrix.
    Theta_ij = self._computeAsymptoticCovarianceMatrix(W_nk, N_k, method = uncertainty_method)
    print "Theta_ij (svd-ew) = "
    print Theta_ij

    # Compute estimators and uncertainties.
    Delta_f_ij = zeros([K,K], dtype = float64)
    dDelta_f_ij = zeros([K,K], dtype = float64)

    Delta_u_ij = zeros([K,K], dtype = float64)
    dDelta_u_ij = zeros([K,K], dtype = float64)        

    Delta_s_ij = zeros([K,K], dtype = float64)
    dDelta_s_ij = zeros([K,K], dtype = float64)
    
    for i in range(0,K):
      for j in range(0,K):
        # Compute dimensionless free energy difference and associated uncertainty (Eq. 14 of [1]).
        Delta_f_ij[i,j] = (self.f_k[j] - self.f_k[i])
        dDelta_f_ij[i,j] = math.sqrt( Theta_ij[i,i] + Theta_ij[j,j] - 2.0 * Theta_ij[i,j] )

        # Compute reduced enthalpy difference.
        Delta_u_ij[i,j] = u_i[j] - u_i[i]
        dDelta_u_ij[i,j] = math.sqrt(
          + u_i[i] * Theta_ij[i,i]   * u_i[i] - u_i[i] * Theta_ij[i,j]   * u_i[j] - u_i[i] * Theta_ij[i,K+i]   * u_i[i] + u_i[i] * Theta_ij[i,K+j]   * u_i[j]
          - u_i[j] * Theta_ij[j,i]   * u_i[i] + u_i[j] * Theta_ij[j,j]   * u_i[j] + u_i[j] * Theta_ij[j,K+i]   * u_i[i] - u_i[j] * Theta_ij[j,K+j]   * u_i[j]
          - u_i[i] * Theta_ij[K+i,i] * u_i[i] + u_i[i] * Theta_ij[K+i,j] * u_i[j] + u_i[i] * Theta_ij[K+i,K+i] * u_i[i] - u_i[i] * Theta_ij[K+i,K+j] * u_i[j]
          + u_i[j] * Theta_ij[K+j,i] * u_i[i] - u_i[j] * Theta_ij[K+j,j] * u_i[j] - u_i[j] * Theta_ij[K+j,K+i] * u_i[i] + u_i[j] * Theta_ij[K+j,K+j] * u_i[j]
          )

        # Compute reduced entropy difference.
        Delta_s_ij[i,j] = Delta_u_ij[i,j] - Delta_f_ij[i,j]
        dDelta_s_ij[i,j] = math.sqrt(
          +  (u_i[i]-1) * Theta_ij[i,i]   * (u_i[i]-1) +  (u_i[i]-1) * Theta_ij[i,j]   * (-u_i[j]+1) +  (u_i[i]-1) * Theta_ij[i,K+i]   * (-u_i[i]) +  (u_i[i]-1) * Theta_ij[i,K+j]   * u_i[j]
          + (-u_i[j]+1) * Theta_ij[j,i]   * (u_i[i]-1) + (-u_i[j]+1) * Theta_ij[j,j]   * (-u_i[j]+1) + (-u_i[j]+1) * Theta_ij[j,K+i]   * (-u_i[i]) + (-u_i[j]+1) * Theta_ij[j,K+j]   * u_i[j]
          + (-u_i[i])   * Theta_ij[K+i,i] * (u_i[i]-1) + (-u_i[i])   * Theta_ij[K+i,j] * (-u_i[j]+1) + (-u_i[i])   * Theta_ij[K+i,K+i] * (-u_i[i]) + (-u_i[i])   * Theta_ij[K+i,K+j] * u_i[j]
          +   u_i[j]    * Theta_ij[K+j,i] * (u_i[i]-1) +   u_i[j]    * Theta_ij[K+j,j] * (-u_i[j]+1) +   u_i[j]    * Theta_ij[K+j,K+i] * (-u_i[i]) +   u_i[j]    * Theta_ij[K+j,K+j] * u_i[j]
          )
        
    # Return expectations and uncertainties.
    return (Delta_f_ij, dDelta_f_ij, Delta_u_ij, dDelta_u_ij, Delta_s_ij, dDelta_s_ij)

  def computePMF(self, u_kn, bin_kn, nbins):
    """Compute the free energy of occupying a number of bins.

    REQUIRED ARGUMENTS
      u_kn[k,n] is the reduced potential energy of snapshot n of state k for which the PMF is to be computed.
      bin_kn[k,n] is the bin index of snapshot n of state k.  bin_kn can assume a value in range(0,nbins)
      nbins is the number of bins

    RETURN VALUES
      f_i[i], i = 0..nbins - the dimensionless free energy of state i, relative to the state of lowest free energy
      df_i[i] is the uncertainty in the difference of f_i with respect to the state of lowest free energy

    NOTES
      This method works by computing the free energy of localizing the system to each bin for the given potential by aggregating the log weights for the given potential.
      To estimate uncertainties, the NxK weight matrix W_nk is augmented to be Nx(K+nbins) in order to accomodate the normalized weights of states where
      the potential is given by u_kn within each bin and infinite potential outside the bin.  The uncertainties with respect to the bin of lowest free energy
      are then computed in the standard way.
    
    """

    K = self.K

    # Compute unnormalized log weights for the given reduced potential u_kn.
    log_w_kn = self._computeUnnormalizedLogWeights(u_kn)
    # Unroll to n-indices
    log_w_n = log_w_kn[self.indices]

    # Compute the free energies for these states.
    f_i = zeros([nbins], float64)
    df_i = zeros([nbins], float64)    
    for i in range(nbins):
      # Get linear n-indices of samples that fall in this bin.
      indices = where(bin_kn[self.indices] == i)[0]

      # Compute dimensionless free energy of occupying state i.
      f_i[i] = - self._logSum( log_w_n[indices] )

    # Compute uncertainties by forming matrix of W_nk.
    N_k = zeros([self.K + nbins], int32)
    N_k[0:K] = self.N_k
    W_nk = zeros([self.N, self.K + nbins], float64)
    W_nk[:,0:K] = self.W_nk
    for i in range(nbins):
      # Get indices of samples that fall in this bin.
      indices = where(bin_kn[self.indices] == i)[0]
      
      # Compute normalized weights for this state.      
      W_nk[indices,K+i] = exp(log_w_n[indices] + f_i[i])

    # Compute asymptotic covariance matrix using specified method.
    Theta_ij = self._computeAsymptoticCovarianceMatrix(W_nk, N_k)
        
    # Determine bin index with lowest free energy.
    j = f_i.argmin()

    # Compute uncertainties with respect to difference in free energy from this state j.
    for i in range(nbins):
      df_i[i] = math.sqrt( Theta_ij[K+i,K+i] + Theta_ij[K+j,K+j] - 2.0 * Theta_ij[K+i,K+j] )

    # Shift free energies so that state j has zero free energy.
    f_i -= f_i[j]

    # Return dimensionless free energy and uncertainty.
    return (f_i, df_i)

  #=============================================================================================
  # PRIVATE METHODS
  #=============================================================================================

  def _pseudoinverse(self, A, tol = 1.0e-10):
    """Compute the Moore-Penrose pseudoinverse.

    REQUIRED ARGUMENTS
      A - the square matrix whose pseudoinverse is to be computed

    RETURN VALUES
      Ainv - the pseudoinverse

    OPTIONAL VALUES
      tol - the tolerance below which singular values are to not be inverted

    NOTES
      This implementation is provided because the 'pinv' function of numpy is broken in the version we were using.

    """

    # Get size
    [M,N] = A.shape
    if N != M:
      raise "pseudoinverse can only be computed for square matrices: dimensions were %d x %d" % (M, N)

    # Make sure A contains no nan.
    if(any(isnan(A))):
      print "A contains nan:"
      print A
      raise ParameterError("A contains nan.")    

    # Compute SVD of A.
    [U, S, Vt] = svd(A)

    # Compute pseudoinverse by taking square root of nonzero singular values.
    Ainv = zeros([M,M], dtype = float64)
    for k in range(M):
      if abs(S[k]) > tol:
        Ainv += (1.0/S[k]) * outer(U[:,k], Vt[k,:])        

    return Ainv
      
  def _computeAsymptoticCovarianceMatrix(self, W, N_k, method = 'svd-ew'):
    """Compute estimator of the asymptotic covariance matrix.
    
    REQUIRED ARGUMENTS    
      W (NxK numarray of Floats) - matrix of normalized weights (see Eq. 6) - W[n,k] is the weight of snapshot n (n = 1..N) in state k
                                      Note that sum(W(:,k)) = 1 for any k = 1..K, and sum(N_k(:) .* W(n,:)) = 1 for any n.
      N_k (K array of int) - N_k[k] is the number of samples from state K

    RETURN VALUES
      Theta (KxK numarray of Floats) - asymptotic covariance matrix (see Eq. 7)

    OPTIONAL ARGUMENTS
      method (string) - choice of method used to compute asymptotic covariance method: 'generalized-inverse', 'svd', 'svd-ew', 'inverse', or 'approximate' (default 'svd-ew')
                        If None is specified, the default is used.

    NOTES
      'generalized-inverse' currently requires computation of the pseudoinverse of an NxN matrix (where N is the total number of samples)
      'svd' computes the generalized inverse using the singular value decomposition -- this should be efficient yet accurate (faster)
      'svd-ev' is the same as 'svd', but uses the eigenvalue decomposition of W'W to bypass the need to perform an SVD (fastest)
      'inverse' only requires standard inversion of a KxK matrix (where K is the number of states), but requires all K states to be different
      'approximate' only requires multiplication of KxN and NxK matrices, but is an approximate underestimate of the uncertainty

    """  

    # Set 'svd-ew' as default if uncertainty method specified as None.
    if method == None:
      method = 'svd-ew'

    # Get dimensions of weight matrix.
    [N,K] = W.shape

    # Check dimensions
    if(K != N_k.size):
      raise ParameterError('W must be NxK, where N_k is a K-dimensional array.')      
    if(sum(N_k) != N):
      raise ParameterError('W must be NxK, where N = sum_k N_k.')
    
    # Check to make sure the weight matrix W is properly normalized.
    tolerance = 1.0e-4 # tolerance for checking equality of sums
#    for k in range(0,K):
#      column_sum = sum(W[:,k])
#      if (abs(column_sum - 1.0) > tolerance):
#        raise ParameterError('Warning: Should have \sum_n W_nk = 1.  Actual sum for state %d was %f' % (k, column_sum))
#    for n in range(0,N):
#      row_sum = sum(W[n,:] * N_k)
#      if (abs(row_sum - 1.0) > tolerance):
#        raise ParameterError('Warning: Should have \sum_k N_k W_nk = 1.  Actual sum of row %d was %f' % (n, row_sum))      

    # Compute estimate of asymptotic covariance matrix using specified method.
    if method == 'generalized-inverse':
      # Use generalized inverse (Eq. 6) -- most general
      # Theta = W' (I - W N W')^+ W

      # Construct matrices
      Ndiag = matrix(diag(N_k), dtype = float64) # Diagonal N_k matrix.
      W = matrix(W, dtype = float64)
      I = identity(N, dtype = float64)

      # Compute covariance
      Theta = W.T * self._pseudoinverse(I - W * Ndiag * W.T) * W      

    elif method == 'inverse':
      # Use standard inverse method (Eq. 8) -- only applicable if all K states are different
      # TODO: Make sure W contains no duplicate columns, or is nonsingular.
      # Theta = [(W'W)^-1 - N + 1 1'/N]^-1
      
      # Construct matrices
      Ndiag = matrix(diag(N_k), dtype = float64) # Diagonal N_k matrix.      
      W = matrix(W, dtype = float64)
      I = identity(N, dtype = float64)
      O = ones([K,K], dtype = float64) / float(N) # matrix of ones, times 1/N

      # Make sure W is nonsingular.
      if (abs(det(W.T * W)) < tolerance):
        print "Warning: W appears to be singular, yet 'inverse' method of uncertainty estimation requires W contain no duplicate states."
    
      # Compute covariance
      Theta = ( (W.T * W).I - Ndiag + O).I

    elif method == 'approximate':
      # Use fast approximate expression -- this underestimates the true covariance, but may be a good approximation in some cases and requires no matrix inversions
      # Theta = P'P

      # Construct matrices
      W = matrix(W, dtype = float64)

      # Compute covariance
      Theta = W.T * W

    elif method == 'svd':
      # Use singular value decomposition based approach given in supplementary material to efficiently compute uncertainty

      # Construct matrices
      Ndiag = matrix(diag(N_k), dtype = float64)
      W = matrix(W, dtype = float64)
      I = identity(K, dtype = float64)      
      
      # Compute SVD of W
      [U, S, Vt] = svd(W)      
      Sigma = matrix(diag(S))
      V = matrix(Vt).T

      # Compute covariance
      Theta = V * Sigma * self._pseudoinverse(I - Sigma * V.T * Ndiag * V * Sigma) * Sigma * V.T

    elif method == 'svd-ew':
      # Use singular value decomposition based approach given in supplementary material to efficiently compute uncertainty
      # The eigenvalue decomposition of W'W is used to forego computing the SVD.

      # Construct matrices
      Ndiag = matrix(diag(N_k), dtype = float64)
      W = matrix(W, dtype = float64)
      I = identity(K, dtype = float64)      
      
      # Compute singular values and right singular vectors of W without using SVD
      # Instead, we compute eigenvalues and eigenvectors of W'W.
      # Note W'W = (U S V')'(U S V') = V S' U' U S V' = V (S'S) V'      
      [S2, V] = eigh(W.T * W)
      # Set any slightly negative eigenvalues to zero.
      S2[where(S2 < 0.0)] = 0.0
      # Form matrix of singular values Sigma, and V.
      Sigma = matrix(diag(sqrt(S2)))
      V = matrix(V)

      # Compute covariance
      Theta = V * Sigma * self._pseudoinverse(I - Sigma * V.T * Ndiag * V * Sigma) * Sigma * V.T

    else:
      # Raise an exception.
      raise ParameterEror('Method ' + method + ' unrecognized.')

    return Theta
      
  def _computeUnnormalizedLogWeights(self, u_kn):
    """Return unnormalized log weights.

    REQUIRED ARGUMENTS
      u_kn (K x N_max numarray Float matrix) - reduced potential energies

    RETURN VALUES
      log_w_kn (K x N_max numarray Float matix) - unnormalized log weights

    TODO
      This function could be speeded significantly through use of compiled code, such as f2py with SciPy.
    """

    # Compute unnormalized log weights.
    log_w_kn = zeros([self.K,self.N_max], dtype = float64)

    for k in range(0,self.K):
      for n in range(0,self.N_k[k]):
        #log_w_kn[k,n] = - u_kn[k,n] - self._logSum( log(self.N_k[self.nonzero_N_k_indices]) + self.f_k[self.nonzero_N_k_indices] - self.u_kln[k,self.nonzero_N_k_indices,n] )
        log_w_kn[k,n] = - self._logSum( log(self.N_k[self.nonzero_N_k_indices]) + self.f_k[self.nonzero_N_k_indices] - (self.u_kln[k,self.nonzero_N_k_indices,n] - u_kn[k,n]) )        

    return log_w_kn

  def _logSum(self, log_terms):
    """Compute the log of a sum of terms whose logarithms are provided.
    
    REQUIRED ARGUMENTS  
      log_terms is the array (possibly multidimensional) containing the logs of the terms to be summed.
  
    RETURN VALUES
      log_sum is the log of the sum of the terms.

    """

    # compute the maximum argument
    max_log_term = log_terms.max()
    
    # compute the reduced terms
    terms = exp(log_terms - max_log_term)

    # compute the log sum
    log_sum = log( terms.sum() ) + max_log_term

    # return the log sum
    return log_sum

  def selfConsistentIteration(self, relative_tolerance = 1.0e-6, maximumIterations = 1000, verbose = True):
    """Determine free energies by self-consistent iteration.

    OPTIONAL ARGUMENTS
      relative_tolerance (float between 0 and 1) - relative tolerance for convergence (default 1.0e-5)
      maximumIterations (int) - maximum number of self-consistent iterations (default 1000)
      verbose (boolean) - verbosity level for debug output

    NOTES
      Self-consistent iteration of the MBAR equations is used, as described in [1].

    REFERENCES
    """

    K = self.K
  
    # Iteratively update dimensionless free energies until convergence to specified tolerance, or maximum allowed number of iterations has been exceeded.
    if verbose: print "MBAR: Computing dimensionless free energies by iteration.  This may take from seconds to minutes, depending on the quantity of data..."
    for iteration in range(0,maximumIterations):
      if verbose: print 'Self-consistent iteration %d' % iteration

      # Store for new estimate of dimensionless relative free energies.
      f_k_new = self.f_k.copy() 

      # Compute updated estimates of dimensionless free energies by Eq. 13.
      for l in range(0,K):
          # Compute unnormalized log weights.
          log_w_kn = self._computeUnnormalizedLogWeights(self.u_kln[:,l,:])
          
          # Update dimensionless free energies.
          f_k_new[l] = - self._logSum( log_w_kn[self.indices] )

      # Shift the dimensionless free energies such that f_k[0] = 0 (to ensure unique solution, and prevent overflow during early iterations).
      f_k_new[:] = f_k_new[:] - f_k_new[0]

      # Compute change from old to new estimate.
      Delta_f_k = f_k_new - self.f_k

      # Update stored free energies.
      self.f_k = f_k_new.copy()

      # Check convergence criteria.
      # Terminate when max((f - fold) / f) < relative_tolerance for all nonzero f.
      max_delta = max(abs(Delta_f_k) / max(abs(f_k_new)))
      if (max_delta < relative_tolerance):
        break

      # write out current estimate
      if verbose:
        print "current f_k ="
        print self.f_k
        print "relative max_delta = %e" % max_delta

    if (maximumIterations == 1):
      return

    # Report convergence, or warn user if convergence was not achieved.
    if (max_delta < relative_tolerance):
      # Convergence achieved.
      if verbose: print 'Converged to tolerance of %e in %d iterations.' % (max_delta, iteration+1)
    else:
      # Warn that convergence was not achieved.
      print 'WARNING: Did not converge to within specified tolerance.'
      print 'max_delta = %e, TOLERANCE = %e, MAX_ITS = %d' % (max_delta, relative_tolerance, maximumIterations)
      
    return

  def NewtonRaphson(self, first_gamma = 0.1, gamma = 1.0, relative_tolerance = 1.0e-6, maximumIterations = 1000, verbose = True):
    """Determine dimensionless free energies by Newton-Raphson iteration.

    OPTIONAL ARGUMENTS
      first_gamma (float between 0 and 1) - step size multiplier to use for first step (default 0.1)
      gamma (float between 0 and 1) - step size multiplier for subsequent steps (default 1.0)
      relative_tolerance (float between 0 and 1) - relative tolerance for convergence (default 1.0e-6)
      maximumIterations (int) - maximum number of Newton-Raphson iterations (default 1000)
      verbose (boolean) - verbosity level for debug output

    NOTES
      This method determines the dimensionless free energies by minimizing a convex function whose solution is the desired estimator.      
      The original idea came from the construction of a likelihood function that independently reproduced the work of Geyer (see [1]
      and Section 6 of [2]).
      This can alternatively be formulated as a root-finding algorithm for the Z-estimator.
      More details of this procedure will follow in a subsequent paper.
      Only those states with nonzero counts are include in the estimation procedure.
      This algorithm is expected to scale poorly with the number of states due to the inversion of the Hessian.

    REFERENCES
      [1] Geyer CJ. Estimating normalizing constants and reweighting mixtures in Markov chain Monte Carlo.
      Technical Report No 568, School of Statistics, University of Minnesota, 1991-1994.
      [2] Kong A, McCullagh P, Meng X-L, Nicolae D, and Tan Z. A theory of statistical models for Monte Carlo integration.
      J. R. Stat Soc B, 65(3):585-618, 2003.
  
    """

    if verbose: print "Determining dimensionless free energies by Newton-Raphson iteration."

    # Number of states with samples.
    K = self.nonzero_N_k_indices.size
    print "There are %d states with samples." % K

    # Free energies
    f_k = self.f_k[self.nonzero_N_k_indices].copy()

    # Samples
    N_k = self.N_k[self.nonzero_N_k_indices].copy()

    # Perform Newton-Raphson iterations
    p_nk = zeros([self.N, K], dtype = float64)    
    for iteration in range(0, maximumIterations):
      if verbose: print "Newton-Raphson iteration %d" % iteration
      
      # Store for new estimate of dimensionless relative free energies.
      # Only dimensionless free energies for the states with samples are stored.
      f_k_new = f_k.copy()

      # Compute p(k | x_n) to allow rapid gradient calculations.
      # TODO: Only use states with N_k > 0 for this and Hessian steps.
      for l in range(0, K):
          # Compute unnormalized log weights.
          log_w_kn = self._computeUnnormalizedLogWeights(self.u_kln[:,self.nonzero_N_k_indices[l],:])

          # Compute p(k | x_n)
          p_nk[:,l] = N_k[l] * exp(log_w_kn[self.indices] + f_k[l])
#          print p_nk[:,l]

      # Compute gradient and Hessian of last (K-1) states.
      # NOTE: Calculation of the gradient and Hessian could be further optimized.
      g = matrix(N_k[1:K] - p_nk[:,1:K].sum(0)).T
      H = matrix(zeros([K-1,K-1], dtype = float64))
      for i in range(1,K):
        H[i-1,i-1] = - (p_nk[:,i] * (1.0 - p_nk[:,i])).sum()
#        print H[i-1,i-1]
        for j in range(1,i):
          H[i-1,j-1] = (p_nk[:,i] * p_nk[:,j]).sum()
          H[j-1,i-1] = H[i-1,j-1]

#      print "g = "
#      print g
#      print "H = "
#      print H
          
      # Update the free energy estimate.
      # The pseudoinverse is used instead of the matrix inverse because H may have less than full rank.
      #Hinvg = H.I * g
      Hinvg = self._pseudoinverse(H) * g
#      print "Hinvg = "
#      print Hinvg
      for k in range(0,K-1):
        if iteration == 0:
          f_k_new[k+1] -= first_gamma * Hinvg[k]
        else:
          f_k_new[k+1] -= gamma * Hinvg[k]          

      # Compute change from old to new estimate.
      Delta_f_k = f_k_new - f_k

      # Update stored free energies.
      f_k = f_k_new.copy()
      self.f_k[self.nonzero_N_k_indices] = f_k

      # compute relative change
      max_delta = max(abs(Delta_f_k) / max(abs(f_k_new)))

      # write out current estimate
      if verbose:
        print "current f_k for states with samples ="
        # censor out elements of f_k that are nonsensical because N_k = 0
        for k in range(0,K):
          print ("%8.4f " % self.f_k[k])

        print "relative max_delta = %e" % max_delta

      # Check convergence criteria.
      # Terminate when max((f - fold) / f) < relative_tolerance for all nonzero f.
      if (max_delta < relative_tolerance):
        break

    # Report convergence, or warn user if convergence was not achieved.
    if (max_delta < relative_tolerance):
      # Convergence achieved.
      if verbose: print 'Converged to tolerance of %e in %d Newton-Raphson iterations.' % (max_delta, iteration+1)
    else:
      # Warn that convergence was not achieved.
      print 'WARNING: Did not converge to within specified tolerance.'
      print 'max_delta = %e, TOLERANCE = %e, MAX_ITS = %d' % (max_delta, relative_tolerance, maximumIterations)

    # Recompute all free energies because those from states with zero samples are not correctly computed by Newton-Raphson.
    if verbose: print "Recomputing all free energies..."
    for l in range(0,K):
      # Compute unnormalized log weights.
      log_w_kn = self._computeUnnormalizedLogWeights(self.u_kln[:,l,:])        
      # Update dimensionless free energies.
      self.f_k[l] = - self._logSum( log_w_kn[self.indices] )
    self.f_k = self.f_k - self.f_k[0]
    
    return
